@@ -16,6 +16,17 @@ server_is_quite = True
 
 from loadlib import Client64
 
+#*** Random String Generator ***
+# Allows us to get a random string to use as the encryption key every time
+import random
+import string
+
+#Function to get random string to use as encryption key
+#Returns a string of N length; containing ASCII uppercase letters, ASCII lower case letters, and numbers
+def get_random_string(string_length=16):
+    random_string = ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits + string.ascii_lowercase) for _ in range(string_length))
+    return str(random_string)
+
 ###############################################
 #*** How to Use ***
 ###############################################
@@ -31,11 +42,33 @@ from loadlib import Client64
 #*** Server Class ***
 ###############################################
 
-class MagicDAQClient64Bit(Client64):
+# We make one instance of the Client64 object and share it among all of the MagicDAQClient64Bit objects (multiple DAQs)
+client64_object = None
 
-    def __init__(self):
+# Make sure we only instantiate once on file load
+if client64_object is None:
+    client64_object = Client64(module32='magicdaq_server_32bit', quiet=server_is_quite)
+
+
+class MagicDAQClient64Bit():
+
+    def __init__(self, local_64_bit_client = client64_object):
         # Specify the name of the Python module to execute on the 32-bit server (i.e., 'my_server')
-        super(MagicDAQClient64Bit, self).__init__(module32='magicdaq_server_32bit', quiet=server_is_quite)
+
+        # Code that is currently deployed in MagicDAQ API::
+        #super(MagicDAQClient64Bit, self).__init__(module32='magicdaq_server_32bit', quiet=server_is_quite)
+
+        # Trying something a bit simpler
+        # https://www.i2tutorials.com/python-super-with-__init__-methods/#:~:text=It%20is%20known%20as%20a,the%20attributes%20of%20the%20class.&text=The%20super()%20function%20allows,the%20base%20class%20name%20explicitly.
+        #super().__init__(module32='magicdaq_server_32bit', quiet=server_is_quite)
+
+        self.local_client = local_64_bit_client
+
+        # Instruct 32bit server to instantiate a DAQ object
+        # Unique ID to identify DAQ object
+        self.unique_daq_id = get_random_string(string_length=32)
+        self.instantiate_daq_object()
+
 
     ###################################################
     # *** All PUBLIC DAQDevice Methods ***
@@ -45,6 +78,12 @@ class MagicDAQClient64Bit(Client64):
     #*** easydaq.py Public Methods ***
     ###################################################
 
+    #*** DAQ Object Instantiation Method ***
+    # Instructs 32bit server to make a new DAQ object
+    def instantiate_daq_object(self):
+        return self.local_client.request32('instantiate_daq_object', self.unique_daq_id)
+
+
     def open_daq_device(self, daq_serial_num = None):
         '''
         Method opens a daq device for utilization.
@@ -63,7 +102,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.open_daq_device(daq_serial_num = '7e18aa9b')
         '''
 
-        return self.request32('open_daq_device', daq_serial_num)
+        return self.local_client.request32('open_daq_device', self.unique_daq_id, daq_serial_num)
 
     def get_hardware_version(self) -> str:
         """
@@ -80,7 +119,7 @@ class MagicDAQClient64Bit(Client64):
         This is Hardware Version: 1.10
         """
 
-        return self.request32('get_hardware_version')
+        return self.local_client.request32('get_hardware_version', self.unique_daq_id)
 
     def get_serial_number(self) -> str:
         """
@@ -97,7 +136,7 @@ class MagicDAQClient64Bit(Client64):
         This is Serial Number: 1.10
         """
 
-        return self.request32('get_serial_number')
+        return self.local_client.request32('get_serial_number', self.unique_daq_id)
 
     def get_firmware_version(self) -> str:
         """
@@ -114,7 +153,7 @@ class MagicDAQClient64Bit(Client64):
         This is Firmware Version: 2.6
         """
 
-        return self.request32('get_firmware_version')
+        return self.local_client.request32('get_firmware_version', self.unique_daq_id)
 
     def get_bootloader_version(self) -> str:
         """
@@ -130,7 +169,7 @@ class MagicDAQClient64Bit(Client64):
         ***Example Usage Output ***
         This is Bootloader Version: 1.0
         """
-        return self.request32('get_bootloader_version')
+        return self.local_client.request32('get_bootloader_version', self.unique_daq_id)
 
     def get_product_type(self) -> str:
         '''
@@ -147,7 +186,7 @@ class MagicDAQClient64Bit(Client64):
         This is Product Type: 3
         '''
 
-        return self.request32('get_product_type')
+        return self.local_client.request32('get_product_type', self.unique_daq_id)
 
     def get_all_version_info(self):
         '''
@@ -172,7 +211,7 @@ class MagicDAQClient64Bit(Client64):
         [['Hardware Version', '1.10'], ['Serial Number', '7e18aa9b'], ['Firmware Version', '2.6'], ['Bootloader Version', '1.0'], ['Product Type', '3']]
         '''
 
-        return self.request32('get_all_version_info')
+        return self.local_client.request32('get_all_version_info', self.unique_daq_id)
 
     def close_daq_device(self):
         '''
@@ -193,7 +232,7 @@ class MagicDAQClient64Bit(Client64):
 
         '''
 
-        return self.request32('close_daq_device')
+        return self.local_client.request32('close_daq_device', self.unique_daq_id)
 
     #*** Former Functions, Now Methods of DAQDevice Class **
 
@@ -213,7 +252,7 @@ class MagicDAQClient64Bit(Client64):
         1.0
         '''
 
-        return self.request32('get_api_version')
+        return self.local_client.request32('get_api_version', self.unique_daq_id)
 
     def get_driver_version(self) -> float:
         '''
@@ -231,7 +270,7 @@ class MagicDAQClient64Bit(Client64):
         1.0
         '''
 
-        return self.request32('get_driver_version')
+        return self.local_client.request32('get_driver_version', self.unique_daq_id)
 
     def list_all_daqs(self):
         '''
@@ -243,7 +282,7 @@ class MagicDAQClient64Bit(Client64):
         print('All DAQ serial numbers: ',easydaq.list_all_daqs())
         '''
 
-        return self.request32('list_all_daqs')
+        return self.local_client.request32('list_all_daqs', self.unique_daq_id)
 
 
     ###################################################
@@ -268,7 +307,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.set_digital_output(0,1)
         '''
 
-        return self.request32('set_digital_output', channel, pin_state)
+        return self.local_client.request32('set_digital_output',self.unique_daq_id, channel, pin_state)
 
     def configure_counter_pwm(self, pwm_frequency: float, pwm_duty_cycle: float, total_cycle_count = 0):
         '''
@@ -310,7 +349,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.start_counter_pwm()
         '''
 
-        return self.request32('configure_counter_pwm', pwm_frequency, pwm_duty_cycle, total_cycle_count)
+        return self.local_client.request32('configure_counter_pwm',self.unique_daq_id, pwm_frequency, pwm_duty_cycle, total_cycle_count)
 
     def start_counter_pwm(self):
         '''
@@ -335,7 +374,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.start_counter_pwm()
         '''
 
-        return self.request32('start_counter_pwm')
+        return self.local_client.request32('start_counter_pwm', self.unique_daq_id)
 
     def stop_counter_pwm(self):
         '''
@@ -368,7 +407,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.stop_counter_pwm()
         '''
 
-        return self.request32('stop_counter_pwm')
+        return self.local_client.request32('stop_counter_pwm', self.unique_daq_id)
 
     def read_digital_input(self, channel: int):
         '''
@@ -395,7 +434,7 @@ class MagicDAQClient64Bit(Client64):
         print("This is pin 0 state: ", daq_one.read_digital_input(0))
         '''
 
-        return self.request32('read_digital_input', channel)
+        return self.local_client.request32('read_digital_input',self.unique_daq_id, channel)
 
     def enable_pulse_counter(self, edge_type = 0):
         '''
@@ -421,7 +460,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.enable_pulse_counter(edge_type = 1)
         '''
 
-        return self.request32('enable_pulse_counter', edge_type)
+        return self.local_client.request32('enable_pulse_counter',self.unique_daq_id, edge_type)
 
     def read_pulse_counter(self) -> int:
         '''
@@ -445,7 +484,7 @@ class MagicDAQClient64Bit(Client64):
         print('This many pulses have been counted: ',daq_one.read_pulse_counter())
         '''
 
-        return self.request32('read_pulse_counter')
+        return self.local_client.request32('read_pulse_counter', self.unique_daq_id)
 
     def clear_pulse_counter(self):
         '''
@@ -466,7 +505,7 @@ class MagicDAQClient64Bit(Client64):
         print('The pulse counter has now been cleared. New pulse count value: ',daq_one_read_pulse_counter())
         '''
 
-        return self.request32('clear_pulse_counter')
+        return self.local_client.request32('clear_pulse_counter', self.unique_daq_id)
 
 
     ###################################################
@@ -498,7 +537,7 @@ class MagicDAQClient64Bit(Client64):
         print('This is voltage at analog pin 0: ',read_single_analog_input(0))
         '''
 
-        return self.request32('read_analog_input', channel, decimal_places)
+        return self.local_client.request32('read_analog_input',self.unique_daq_id, channel, decimal_places)
 
     def read_diff_analog_input(self, channel_p: int, channel_n: int, decimal_places = 2) -> float:
         '''
@@ -526,7 +565,7 @@ class MagicDAQClient64Bit(Client64):
         print('This is the differential voltage between pin 1 and pin 0: ',read_raw_diff_analog_input(1,0))
         '''
 
-        return self.request32('read_diff_analog_input', channel_p, channel_n, decimal_places)
+        return self.local_client.request32('read_diff_analog_input',self.unique_daq_id, channel_p, channel_n, decimal_places)
 
     def set_analog_output(self, channel: int, output_voltage: float):
         '''
@@ -546,7 +585,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.set_digital_output(0,1)
         '''
 
-        return self.request32('set_analog_output', channel, output_voltage)
+        return self.local_client.request32('set_analog_output',self.unique_daq_id, channel, output_voltage)
 
     def configure_analog_output_sine_wave(self, channel: int, sine_frequency: float, total_cycle_count=0, amplitude=5):
         '''
@@ -583,7 +622,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.stop_analog_output_wave(1)
         '''
 
-        return self.request32('configure_analog_output_sine_wave', channel, sine_frequency, total_cycle_count, amplitude)
+        return self.local_client.request32('configure_analog_output_sine_wave',self.unique_daq_id, channel, sine_frequency, total_cycle_count, amplitude)
 
     def configure_analog_output_pwm_wave(self, channel: int, pwm_frequency: float, pwm_duty_cycle: float, total_cycle_count=0, amplitude=5):
         '''
@@ -622,7 +661,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.stop_analog_output_wave(0)
         '''
 
-        return self.request32('configure_analog_output_pwm_wave',channel,pwm_frequency,pwm_duty_cycle,total_cycle_count,amplitude)
+        return self.local_client.request32('configure_analog_output_pwm_wave',self.unique_daq_id, channel,pwm_frequency,pwm_duty_cycle,total_cycle_count,amplitude)
 
     def start_analog_output_wave(self, channel: int):
         '''
@@ -654,7 +693,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.stop_analog_output_wave(1)
         '''
 
-        return self.request32('start_analog_output_wave', channel)
+        return self.local_client.request32('start_analog_output_wave',self.unique_daq_id, channel)
 
     def stop_analog_output_wave(self, channel: int):
         '''
@@ -682,7 +721,7 @@ class MagicDAQClient64Bit(Client64):
         daq_one.stop_analog_output_wave(1)
         '''
 
-        return self.request32('stop_analog_output_wave', channel)
+        return self.local_client.request32('stop_analog_output_wave',self.unique_daq_id, channel)
 
 
     ###################################################
@@ -708,7 +747,7 @@ class MagicDAQClient64Bit(Client64):
         TODO: complete
         '''
 
-        return self.request32('configure_analog_input_stream', channels, measurement_frequency, decimal_places)
+        return self.local_client.request32('configure_analog_input_stream',self.unique_daq_id, channels, measurement_frequency, decimal_places)
 
     def start_analog_input_stream(self) -> float:
         '''
@@ -722,7 +761,7 @@ class MagicDAQClient64Bit(Client64):
         TODO: complete
         '''
 
-        return self.request32('start_analog_input_stream')
+        return self.local_client.request32('start_analog_input_stream', self.unique_daq_id)
 
     def set_streaming_data_buffer_max_rows(self, max_rows: int):
         """
@@ -738,7 +777,7 @@ class MagicDAQClient64Bit(Client64):
         #TODO: put something here
         """
 
-        return self.request32('set_streaming_data_buffer_max_rows', max_rows)
+        return self.local_client.request32('set_streaming_data_buffer_max_rows',self.unique_daq_id, max_rows)
 
     def get_last_n_streaming_data_samples(self, n_samples:int):
         '''
@@ -755,7 +794,7 @@ class MagicDAQClient64Bit(Client64):
         TODO: complete
         '''
 
-        return self.request32('get_last_n_streaming_data_samples', n_samples)
+        return self.local_client.request32('get_last_n_streaming_data_samples',self.unique_daq_id, n_samples)
 
     def get_full_streaming_data_buffer(self, only_new_data = True, read_and_delete =False):
         '''
@@ -777,7 +816,7 @@ class MagicDAQClient64Bit(Client64):
         TODO: complete
         '''
 
-        return self.request32('get_full_streaming_data_buffer', only_new_data, read_and_delete)
+        return self.local_client.request32('get_full_streaming_data_buffer',self.unique_daq_id, only_new_data, read_and_delete)
 
     def delete_streaming_data_buffer(self):
         '''
@@ -791,7 +830,7 @@ class MagicDAQClient64Bit(Client64):
         TODO: complete
         '''
 
-        return self.request32('delete_streaming_data_buffer')
+        return self.local_client.request32('delete_streaming_data_buffer', self.unique_daq_id)
 
     def stop_analog_input_stream(self):
         '''
@@ -805,4 +844,4 @@ class MagicDAQClient64Bit(Client64):
         TODO: complete
         '''
 
-        return self.request32('stop_analog_input_stream')
\ No newline at end of file
+        return self.local_client.request32('stop_analog_input_stream', self.unique_daq_id)
\ No newline at end of file
